import asyncio
import logging

from trimble.id._constants import PACKAGE_NAME
from .http_client import HttpClient
from .http_client import HttpException
from .AnalyticsHttpClient import AnalyticsHttpClient
import pkg_resources  
import asyncio

_logger = logging.getLogger(__name__)
_logger.addHandler(logging.NullHandler())


class RobustHttpClientProvider:
    """
    A HttpClient provider to make HTTP connections with protection against transient failures.
    """
    def __init__(self, httpClientProvider, retries, backoff, productName = None):
        """
        Initialize RobustHttpClientProvider class

        :param httpClientProvider: A HttpClient provider for APIs using Bearer token authorization
        :param retries: The number of retries to attempt a failed request
        :param backoff: The amount of time to back off before a request retry
        :param productName: The produce name of the consuming application
        """

        self._httpClientProvider = httpClientProvider
        self._retries = retries
        self._backoff = backoff
        self._client = None
        self._version = pkg_resources.require(PACKAGE_NAME)[0].version

        AnalyticsHttpClient.SendInitEvent(
            name=self.__class__.__name__, 
            client_name=PACKAGE_NAME, 
            client_version=self._version)

    async def retrieve_client(self):
        """
        Retrieves robust http client that helps to protect against transient failures.
        """

        AnalyticsHttpClient.SendMethodEvent(
            name=f"{self.__class__.__name__}_retrieve_client", 
            client_name=PACKAGE_NAME, 
            client_version=self._version)

        if not self._client:
            self._client = self._RobustHttpClient(
                await self._httpClientProvider.retrieve_client(),
                self._retries,
                self._backoff,
            )

        return self._client

    class _RobustHttpClient:

        # retry for the following status codes
        # 408 Request timeout
        # 409 Conflict
        # 418 I'm a teapot
        # 429 Too many requests
        # 500 Internal server error
        # 502 Bad gateway
        # 503 Service unavailable
        # 504 Gateway timeout
        SUPPORTED_EXCEPTIONS = [408, 409, 418, 429, 500, 502, 503, 504]

        def __init__(self, httpClient, retries=3, backoff=5):
            self._httpClient = httpClient
            self._retries = retries
            self._backoff = backoff

        async def post(self, url, value=None, headers=None):
            """ Http post request """
            headers = headers if headers is not None else {}
            intitialException = None
            retries = -1
            for retries in range(self._retries):
                try:
                    return await self._httpClient.post(url, value, headers)
                except HttpException as ex:
                    if ex.status not in self.SUPPORTED_EXCEPTIONS:
                        _logger.error(
                            "POST %s try %d raised an unsupported exception: %s",
                            url,
                            retries,
                            str(ex),
                            exc_info=ex,
                        )
                        raise
                    else:
                        _logger.info(
                            "POST %s try %d raised a supported exception %s",
                            url,
                            retries,
                            str(ex),
                        )

                    if not intitialException:
                        intitialException = ex

                    _logger.warning(
                        "Post try %d of %d to %s failed with retryable exception",
                        retries,
                        self._retries,
                        url,
                        exc_info=ex,
                    )
                    await asyncio.sleep(self._backoff)

            _logger.error(
                "POST %s giving up after try %d. Initial exception was %s",
                url,
                retries,
                str(intitialException),
            )
            raise intitialException

        async def get(self, url, headers=None):
            """ Http get request """
            headers = headers if headers is not None else {}
            intitialException = None
            retries = -1
            for retries in range(self._retries):
                try:
                    return await self._httpClient.get(url, headers)
                except HttpException as ex:
                    if ex.status not in self.SUPPORTED_EXCEPTIONS:
                        _logger.error(
                            "GET %s try %d raised an unsupported exception: %s",
                            url,
                            retries,
                            str(ex),
                            exc_info=ex,
                        )
                        raise
                    else:
                        _logger.info(
                            "GET %s try %d raised a supported exception %s",
                            url,
                            retries,
                            str(ex),
                        )

                    if not intitialException:
                        intitialException = ex

                    await asyncio.sleep(self._backoff)

            _logger.error(
                "GET %s giving up after try %d. Initial exception was %s",
                url,
                retries,
                str(intitialException),
            )
            raise intitialException

        async def put(self, url, value=None, headers=None):
            """ Http put request """
            headers = headers if headers is not None else {}
            intitialException = None
            retries = -1
            for retries in range(self._retries):
                try:
                    return await self._httpClient.put(url, value, headers)
                except HttpException as ex:
                    if ex.status not in self.SUPPORTED_EXCEPTIONS:
                        _logger.error(
                            "PUT %s try %d raised an unsupported exception: %s",
                            url,
                            retries,
                            str(ex),
                            exc_info=ex,
                        )
                        raise
                    else:
                        _logger.info(
                            "PUT %s try %d raised a supported exception %s",
                            url,
                            retries,
                            str(ex),
                        )

                    if not intitialException:
                        intitialException = ex

                    await asyncio.sleep(self._backoff)

            _logger.error(
                "PUT %s giving up after try %d. Initial exception was %s",
                url,
                retries,
                str(intitialException),
            )
            raise intitialException

        async def patch(self, url, value=None, headers=None):
            """ Http patch request """
            headers = headers if headers is not None else {}
            intitialException = None
            retries = -1
            for retries in range(self._retries):
                try:
                    return await self._httpClient.patch(url, value, headers)
                except HttpException as ex:
                    if ex.status not in self.SUPPORTED_EXCEPTIONS:
                        _logger.error(
                            "PATCH %s try %d raised an unsupported exception: %s",
                            url,
                            retries,
                            str(ex),
                            exc_info=ex,
                        )
                        raise
                    else:
                        _logger.info(
                            "PATCH %s try %d raised a supported exception %s",
                            url,
                            retries,
                            str(ex),
                        )

                    if not intitialException:
                        intitialException = ex

                    await asyncio.sleep(self._backoff)

            _logger.error(
                "PATCH %s giving up after try %d. Initial exception was %s",
                url,
                retries,
                str(intitialException),
            )
            raise intitialException

        async def delete(self, url, headers=None):
            """ delete Http request """
            headers = headers if headers is not None else {}
            intitialException = None
            retries = -1
            for retries in range(self._retries):
                try:
                    return await self._httpClient.delete(url, headers)
                except HttpException as ex:
                    if ex.status not in self.SUPPORTED_EXCEPTIONS:
                        _logger.error(
                            "DELETE %s try %d raised an unsupported exception: %s",
                            url,
                            retries,
                            str(ex),
                            exc_info=ex,
                        )
                        raise
                    else:
                        _logger.info(
                            "DELETE %s try %d raised a supported exception %s",
                            url,
                            retries,
                            str(ex),
                        )

                    if not intitialException:
                        intitialException = ex

                    await asyncio.sleep(self._backoff)

            _logger.error(
                "DELETE %s giving up after try %d. Initial exception was %s",
                url,
                retries,
                str(intitialException),
            )
            raise intitialException

        async def post_json(self, url, value=None, headers=None):
            """ postJSON Http request """
            headers = headers if headers is not None else {}
            intitialException = None
            retries = -1
            for retries in range(self._retries):
                try:
                    return await self._httpClient.postJSON(url, value, headers)
                except HttpException as ex:
                    if ex.status not in self.SUPPORTED_EXCEPTIONS:
                        _logger.error(
                            "POSTjson %s try %d raised an unsupported exception: %s",
                            url,
                            retries,
                            str(ex),
                            exc_info=ex,
                        )
                        raise
                    else:
                        _logger.info(
                            "POSTjson %s try %d raised a supported exception %s",
                            url,
                            retries,
                            str(ex),
                        )

                    if not intitialException:
                        intitialException = ex

                    await asyncio.sleep(self._backoff)

            _logger.error(
                "POSTjson %s giving up after try %d. Initial exception was %s",
                url,
                retries,
                str(intitialException),
            )
            raise intitialException

        async def get_json(self, url, headers=None):
            """ getJSON Http request """
            headers = headers if headers is not None else {}
            intitialException = None
            retries = -1
            for retries in range(self._retries):
                try:
                    return await self._httpClient.getJSON(url, headers)
                except HttpException as ex:
                    if ex.status not in self.SUPPORTED_EXCEPTIONS:
                        _logger.error(
                            "GETjson %s try %d raised an unsupported exception: %s",
                            url,
                            retries,
                            str(ex),
                            exc_info=ex,
                        )
                        raise
                    else:
                        _logger.info(
                            "GETjson %s try %d raised a supported exception %s",
                            url,
                            retries,
                            str(ex),
                        )

                    if not intitialException:
                        intitialException = ex

                    await asyncio.sleep(self._backoff)

            _logger.error(
                "GETjson %s giving up after try %d. Initial exception was %s",
                url,
                retries,
                str(intitialException),
            )
            raise intitialException

        async def put_json(self, url, value=None, headers=None):
            """ putJSON Http request """
            headers = headers if headers is not None else {}
            intitialException = None
            retries = -1
            for retries in range(self._retries):
                try:
                    return await self._httpClient.putJSON(url, value, headers)
                except HttpException as ex:
                    if ex.status not in self.SUPPORTED_EXCEPTIONS:
                        _logger.error(
                            "PUTjson %s try %d raised an unsupported exception: %s",
                            url,
                            retries,
                            str(ex),
                            exc_info=ex,
                        )
                        raise
                    else:
                        _logger.info(
                            "PUTjson %s try %d raised a supported exception %s",
                            url,
                            retries,
                            str(ex),
                        )

                    if not intitialException:
                        intitialException = ex

                    await asyncio.sleep(self._backoff)

            _logger.error(
                "PUTjson %s giving up after try %d. Initial exception was %s",
                url,
                retries,
                str(intitialException),
            )
            raise intitialException

        async def patch_json(self, url, value=None, headers=None):
            """ patchJSON Http request """
            headers = headers if headers is not None else {}
            intitialException = None
            retries = -1
            for retries in range(self._retries):
                try:
                    return await self._httpClient.patchJSON(url, value, headers)
                except HttpException as ex:
                    if ex.status not in self.SUPPORTED_EXCEPTIONS:
                        _logger.error(
                            "PATCHjson %s try %d raised an unsupported exception: %s",
                            url,
                            retries,
                            str(ex),
                            exc_info=ex,
                        )
                        raise
                    else:
                        _logger.info(
                            "PATCHjson %s try %d raised a supported exception %s",
                            url,
                            retries,
                            str(ex),
                        )

                    if not intitialException:
                        intitialException = ex

                    await asyncio.sleep(self._backoff)

            _logger.error(
                "PATCHjson %s giving up after try %d. Initial exception was %s",
                url,
                retries,
                str(intitialException),
            )
            raise intitialException



from trimble.id._constants import PACKAGE_NAME
from .http_client import HttpClient
from .AnalyticsHttpClient import AnalyticsHttpClient
import pkg_resources  

class OpenIdEndpointProvider:
    """
    An endpoint provider that returns values from a OpenID well known configuration
    """
    def __init__(self, configurationUrl, productName = None):
        """
        Initialize OpenIdEndpointProvider class

        :param configurationUrl: The URL for the Trimble Identity OpenID well know configuration endpoint
        :param productName: Product name of consuming application
        """
        self._configurationUrl = configurationUrl
        self._configuration = None

        self._version = pkg_resources.require(PACKAGE_NAME)[0].version

        AnalyticsHttpClient.SendInitEvent(
            name=self.__class__.__name__, 
            client_name=PACKAGE_NAME, 
            client_version=self._version)
    
    async def retrieve_authorization_endpoint(self):
        """
        Retrieves authorization endpoint from a OpenID well known configuration
        """
        AnalyticsHttpClient.SendMethodEvent(
            name=f"{self.__class__.__name__}_retrieve_authorization_endpoint", 
            client_name=PACKAGE_NAME, 
            client_version=self._version)

        if self._configuration is None:
            await self._load_configuration()
        return self._configuration['authorization_endpoint']

    async def retrieve_token_endpoint(self):
        """
        Retrieves token endpoint from a OpenID well known configuration
        """
        AnalyticsHttpClient.SendMethodEvent(
            name=f"{self.__class__.__name__}_retrieve_token_endpoint", 
            client_name=PACKAGE_NAME, 
            client_version=self._version)

        if self._configuration is None:
            await self._load_configuration()
        return self._configuration['token_endpoint']

    async def retrieve_jwks_endpoint(self):
        """
        Retrieves JWKS endpoint from a OpenID well known configuration
        """
        AnalyticsHttpClient.SendMethodEvent(
            name=f"{self.__class__.__name__}_retrieve_jwks_endpoint", 
            client_name=PACKAGE_NAME, 
            client_version=self._version)

        if self._configuration is None:
            await self._load_configuration()
        return self._configuration['jwks_uri']

    async def _load_configuration(self):
        client = HttpClient('', {})
        self._configuration = await client.get_json(self._configurationUrl, {})


from trimble.id._constants import PACKAGE_NAME
from .AnalyticsHttpClient import AnalyticsHttpClient
import pkg_resources  

class FixedEndpointProvider:
    """
    An endpoint provider that returns fixed OAuth endpoints
    """
    def __init__(self, authorizationEndpoint, tokenEndpoint, userInfoEndpoint, tokenRevocationEndpoint = None, jwksEndpoint = None, productName = None):
        """
        Initialize FixedEndpointProvider class

        :param authorizationEndpoint: Set Authorization Endpoint
        :param tokenEndpoint: Set Token Endpoint
        :param userInfoEndpoint: Set UserInfo Endpoint
        :param tokenRevocationEndpoint: Set Token revocation Endpoint
        :param jwksEndpoint: Set JSON Web key set Endpoint
        :param productName: Product name of the consuming application (optional)
        """
        self._authorizationEndpoint = authorizationEndpoint
        self._tokenEndpoint = tokenEndpoint
        self._userInfoEndpoint = userInfoEndpoint
        self._tokenRevocationEndpoint = tokenRevocationEndpoint
        self._jwksEndpoint = jwksEndpoint
        self._version = pkg_resources.require(PACKAGE_NAME)[0].version

        AnalyticsHttpClient.SendInitEvent(
            name=self.__class__.__name__, 
            client_name=PACKAGE_NAME, 
            client_version=self._version)

    async def retrieve_authorization_endpoint(self):
        """
        Retrieves a fixed Authorization endpoint
        """
        AnalyticsHttpClient.SendMethodEvent(
            name=f"{self.__class__.__name__}_retrieve_authorization_endpoint", 
            client_name=PACKAGE_NAME, 
            client_version=self._version)
        return self._authorizationEndpoint

    async def retrieve_token_endpoint(self):
        """
        Retrieves a fixed Token endpoint
        """
        AnalyticsHttpClient.SendMethodEvent(
            name=f"{self.__class__.__name__}_retrieve_token_endpoint", 
            client_name=PACKAGE_NAME, 
            client_version=self._version)
        return self._tokenEndpoint

    async def retrieve_json_web_keyset_endpoint(self):
        """
        Retrieves a fixed JSON Web key set endpoint
        """
        AnalyticsHttpClient.SendMethodEvent(
            name=f"{self.__class__.__name__}_retrieve_json_web_keyset_endpoint", 
            client_name=PACKAGE_NAME, 
            client_version=self._version)
        return self._jwksEndpoint


from trimble.id._constants import PACKAGE_NAME
from .AnalyticsHttpClient import AnalyticsHttpClient
import pkg_resources  

class FixedKeySetProvider:
    """
    A keyset provider that returns a fixed keyset
    """
    def __init__(self, keyset, productName = None):      
        """
        Initialize FixedKeySetProvider class

        :param keyset: A dictionary of named keys
        :param productName: Product name of consuming application
        """  
        self._keyset = keyset

        self._version = pkg_resources.require(PACKAGE_NAME)[0].version

        AnalyticsHttpClient.SendInitEvent(
            name=self.__class__.__name__, 
            client_name=PACKAGE_NAME, 
            client_version=self._version)

    async def retrieve_keyset(self):
        """
        Retrieves a dictionary of named keys

        :return: Fixed Keyset
        """
        AnalyticsHttpClient.SendMethodEvent(
            name=f"{self.__class__.__name__}_retrieve_keyset", 
            client_name=PACKAGE_NAME, 
            client_version=self._version)
        return self._keyset


import json
import jwt
from .AnalyticsHttpClient import AnalyticsHttpClient

from trimble.id._constants import PACKAGE_NAME
from .http_client import HttpClient
import pkg_resources  

class OpenIdKeySetProvider:
    """
    A keyset provider based on the OAuth well known configuration
    """
    def __init__(self, endpointProvider, productName = None):
        """
        Initialize OpenIdKeySetProvider class

        :param endpointProvider: An endpoint provider that provides the URL for the Trimble Identity JSON web keyset endpoint
        :param productName: Product name of the consuming application
        """
        self._endpointProvider = endpointProvider

        self._version = pkg_resources.require(PACKAGE_NAME)[0].version

        AnalyticsHttpClient.SendInitEvent(
            name=self.__class__.__name__, 
            client_name=PACKAGE_NAME, 
            client_version=self._version)

    async def retrieve_keys(self):
        """
        Retrieves an dictionary of named keys
        """

        AnalyticsHttpClient.SendMethodEvent(
            name=f"{self.__class__.__name__}_retrieve_keys", 
            client_name=PACKAGE_NAME, 
            client_version=self._version)

        client = HttpClient('', {})
        result = await client.get_json(await self._endpointProvider.retrieve_jwks_endpoint(), {})
        self._keys = {}
        for jwk in result['keys']:
            kid = jwk['kid']
            self._keys[kid] = jwt.algorithms.RSAAlgorithm.from_jwk(json.dumps(jwk))
        return self._keys


from trimble.id._constants import PACKAGE_NAME
from .http_client import HttpClient
from .AnalyticsHttpClient import AnalyticsHttpClient
import pkg_resources  

class BearerTokenHttpClientProvider:
    """
    A HttpClient provider for APIs using Bearer token authorization
    """
    def __init__(self, tokenProvider, baseAddress, defaultHeaders = {}, productName=None):
        """
        Initialize Bearer token http client provider

        :param tokenProvider: A token provider that provides the access token for the authenticated application or user
        :param baseAddress: The base address for the API that will be called
        :param defaultHeaders: If any default address to be applied in Http request call. (Optional)
        :param produceName: Specify product name of consuming application (optional)
        """
        self._tokenProvider = tokenProvider
        self._baseAddress = baseAddress
        self._defaultHeaders = defaultHeaders
        self._version = pkg_resources.require(PACKAGE_NAME)[0].version

        AnalyticsHttpClient.SendInitEvent(
            name=self.__class__.__name__, 
            client_name=PACKAGE_NAME, 
            client_version=self._version)

    async def retrieve_client(self):
        """
        Retrieves a preconfigured HttpClient to access a given API

        :return: A preconfigured HttpClient to access any given API
        """
        AnalyticsHttpClient.SendMethodEvent(
            name=f"{self.__class__.__name__}_retrieve_client", 
            client_name=PACKAGE_NAME, 
            client_version=self._version)
        
        token = await self._tokenProvider.retrieve_token()
        
        if 'trimblepaas.com' in self._baseAddress:
            url = self._baseAddress + 'api/'
        else:
            url = self._baseAddress

        url = self._add_trailing_slash(url)

        return HttpClient(url, { **self._defaultHeaders, **{ 'authorization': 'Bearer ' + token } })

    def _add_trailing_slash(self, url):
        if url.endswith('/'):
            return url
        return url + '/'



from trimble.id._constants import PACKAGE_NAME
from .AnalyticsHttpClient import AnalyticsHttpClient
import pkg_resources  

class FixedTokenProvider:
    """
    A token provider that returns a fixed token
    """
    def __init__(self, token, productName = None):
        """
        Initialize FixedTokenProvider class

        :param token: Sets access token to FixedTokenProvider
        :param productName: Product name of the consuming application
        """
        self._token = token

        self._version = pkg_resources.require(PACKAGE_NAME)[0].version

        AnalyticsHttpClient.SendInitEvent(
            name=self.__class__.__name__, 
            client_name=PACKAGE_NAME, 
            client_version=self._version)

    async def retrieve_token(self):
        """
        Retrieves fixed token
        """
        AnalyticsHttpClient.SendMethodEvent(
            name=f"{self.__class__.__name__}_retrieve_token", 
            client_name=PACKAGE_NAME, 
            client_version=self._version)

        return self._token


from urllib.parse import urljoin
from enum import Enum
import requests
import uuid

class EventType(Enum):
    INIT = 1
    METHOD_USAGE = 2
    EXCEPTION = 3

class AnalyticsHttpClient:
    """
    A Http client for sending event hits to GA
    """

    _baseAddress = "https://www.google-analytics.com"
    _measurementID = "G-9DHNE789XG"
    _apiKey = "7wd60wEeSnufb8FK6naHOQ"
    _clientID = str(uuid.uuid4())

    @staticmethod
    def SendInitEvent(name, client_name = None, client_version = None, cid = None):
        """
        Send initialization events

        :param name: Name of an event
        :param client_name: Name of the client SDK
        :param client_version: Version of the client SDK
        :param cid: Client ID (optional)
        """
        params = { 
            "language": "python-sdk", 
            "client_name": client_name,
            "client_version": client_version,
            "event_type": EventType.INIT.name
        }
        AnalyticsHttpClient.SendEvent(name, params, cid)

    @staticmethod
    def SendMethodEvent(name, client_name = None, client_version = None, cid = None):
        """
        Send method events

        :param name: Name of an event
        :param client_name: Name of the client SDK
        :param client_version: Version of the client SDK
        :param cid: Client ID (optional)
        """
        params = { 
            "language": "python-sdk", 
            "client_name": client_name,
            "client_version": client_version,
            "event_type": EventType.METHOD_USAGE.name
        }
        AnalyticsHttpClient.SendEvent(name, params, cid)
 
    @staticmethod
    def SendEvent(name, params, cid=None):
        """
        The method to send other type of events

        :param name: Name of an event
        :param params: Additional parameters of an event
        :param cid: Client ID (optional)
        """
        try:
            if cid == None:
                cid = AnalyticsHttpClient._clientID
            url = AnalyticsHttpClient._combineUrl(
                AnalyticsHttpClient._baseAddress, 
                f"/mp/collect?measurement_id={AnalyticsHttpClient._measurementID}&api_secret={AnalyticsHttpClient._apiKey}")
            headers = {
                'User-Agent':'TrimbleCloud.AnalyticsHttpClient/1.1.0',
                'Content-Type':'application/json'
                }
            data = {
                "client_id": cid,
                "events": [
                    {
                        "name": name,
                        "params": params
                    }
                ]
            }
            requests.post(url,headers=headers, json=data)
        except:
            pass

    def _combineUrl(baseAddress, url):
        return urljoin(baseAddress, url)

    def _combineHeaders(baseHeaders, headers):
        return {**baseHeaders, **headers}